#! /usr/bin/perl

# Translate between Forth screens files (*.blk) and text files with
# Forth source code in them (*.fth).

# The program takes an input file, with an extension of .blk or
# .fth. It creates the output file with the same base name and the
# other extension. It then processes the input file, using the process
# indicated by the input extension. For example,

# blk2fth foo.blk

# will produce the text file foo.fth from the forth screens in
# foo.blk. To go the other way,

# blk2fth foo.fth

# Note: This utility was developed by Charles Curley to convert standard
# Forth screens (16 rows x 64 columns) and adapted by Kelly A. Murta for
# use with ZX81 Toddy Forth-79 screens (32 rows x 32 columns). As a result,
# it was renamed from scr2scn.pl to blk2fth.pl. The original is available
# on GitHub: https://github.com/charlescurley/scr2scn

use strict;
use 5.010;
use warnings;
use Getopt::Long qw(:config pass_through);
use File::Basename;

my $iNeedHelp;
my $script = basename ($0);
my $verbose = 0;

sub process () {
  my $blkFile = $_[0];

  if ($blkFile eq "") {
    print "ERROR: The source file is missing!\n";
    help ();
    exit;
  }

  # regex from https://perldoc.perl.org/File/Basename.html
  my ($base, $dir, $ext) = fileparse ($blkFile, qr/\.[^.]*/);

  my $destFile = ${dir} . uc(${base});

  if (uc($ext) eq '.BLK') {
    $destFile .= '.FTH';
  } elsif (uc($ext) eq '.FTH') {
    $destFile .= '.BLK';
  } else {
    help ();
    die ("$ext is not a valid extension.\n")
  }


  if ($verbose) {
    print ("\$base is |$base| \$ext is |$ext|. \$dir is |$dir|\n");
    print ("Source file name is |${dir}${base}${ext}|\n");
    print ("Destination file is |$destFile|\n");
  }

  # open DEST read and write, create & truncate.
  open (DEST, "+> $destFile") or die ("Can't open destination file $destFile.\n");
  open (BLK,  "< $blkFile")  or die ("Can't open source file $blkFile.\n");

  my $buf;
  my @array;
  my $i;

  if (lc($ext) eq ".blk") {
    # screens file to text file.
    while (read (BLK, $buf, 32)) {
      @array = unpack ("C*", $buf);
      if ($array[0] == 0 and $array[1] == 0) {
        # Non-printable screen
        print "Non-printable screen.\n" if ($verbose);
        print DEST "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n";
        my $results = read (BLK, $buf, 1024-32);
        if ($results != 1024-32) {
          die ("Read error in non-printable screen.\n");
        }
      } else {
        # Printable screen

        # first line.....
        $buf =~ s/\s+$//;       # trim trailing spaces.
        print DEST "$buf\n";

        # subsequent lines....
        for ($i = 31; $i > 0 ; $i--) {
          my $result = read (BLK, $buf, 32);
          $buf =~ s/\s+$//;     # trim trailing spaces.
          print DEST "$buf\n";
        }
      }
    }

  } else {
    # File to screens file.
    my $line;

    # 32 blanks, for padding out each line.
    my $blanks = '                                ';

    while (defined ($line = <BLK>)) {
      chomp $line;
      my $size = length ($line);
      print DEST $line;

      print DEST substr ($blanks, 0, 32-$size);

      $i++;
    }

    print ("$i lines shipped so far.\n") if ($verbose);
    while ($i%32 != 0) {
      print ("Shipping new line $i\n") if ($verbose);
      print DEST $blanks;
      $i++;
    }

    # OK, we've made our file. Now find screens with nothing but blanks
    # in them, and stick 0 in the first two words.
    seek (DEST, 0, 0);

    # prepare our two bytes of 0s for blank screens.
    my $blankScreen = '';
    $array[0] = $array[1] = 0;
    $blankScreen = pack ("C*", @array);

    $i = 0;

    while (read (DEST, $buf, 1024)) {

      # print $i % 15 . "\n";

      if (($i % 31) == 0 and $verbose) {
        print sprintf("\n%5d:", $i);
      }

      $buf =~ s/\s+$//;         # trim trailing spaces.

      if (length ($buf) == 0) {
        # print ("$i: Non-priniting screen.\n") if ($verbose);

        # seek back one screen....
        seek (DEST, -1024, 1);

        # write the two 0s already prepared....
        print DEST $blankScreen;

        # seek forward to the next screen....
        seek (DEST, +1022, 1);

        printf ('%4d', $i) if ($verbose);
      } else {
        print '   .' if ($verbose);
      }

      $i++;
    }
  }

  close (BLK)  or die ("Can't close source file $blkFile.\n");
  close (DEST) or die ("Can't close destination file $destFile.\n");

}

sub help () {
  print qq~
Usage:
   $script [OPTIONS] sourceFile
   Valid extensions are .blk and .fth, upper or lower case

Options:
   --help           Show this screen
   --verbose        Be verbose
~;
}


GetOptions("help"      => \$iNeedHelp,
           "verbose"   => \$verbose,
          )
  or die("Error in command line arguments\n");

if ($iNeedHelp) {
  help;
  exit;
}

while (@ARGV) {
  &process (shift @ARGV);
}

exit (0);